from typing import Any, ClassVar, overload

from qiskit import QuantumCircuit

from mqt.qcec.types import ApplicationSchemeName, EquivalenceCriterionName, StateTypeName

class ApplicationScheme:
    __members__: ClassVar[dict[ApplicationScheme, int]] = ...  # read-only
    gate_cost: ClassVar[ApplicationScheme] = ...
    lookahead: ClassVar[ApplicationScheme] = ...
    one_to_one: ClassVar[ApplicationScheme] = ...
    proportional: ClassVar[ApplicationScheme] = ...
    sequential: ClassVar[ApplicationScheme] = ...
    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: ApplicationSchemeName) -> None: ...
    def name(self) -> ApplicationSchemeName: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> ApplicationSchemeName: ...
    @property
    def value(self) -> int: ...

class Configuration:
    class Application:
        alternating_scheme: ApplicationScheme
        construction_scheme: ApplicationScheme
        profile: str
        simulation_scheme: ApplicationScheme
        def __init__(self) -> None: ...

    class Execution:
        nthreads: int
        numerical_tolerance: float
        parallel: bool
        run_alternating_checker: bool
        run_construction_checker: bool
        run_simulation_checker: bool
        run_zx_checker: bool
        timeout: float
        def __init__(self) -> None: ...

    class Functionality:
        trace_threshold: float
        def __init__(self) -> None: ...

    class Optimizations:
        fix_output_permutation_mismatch: bool
        fuse_single_qubit_gates: bool
        reconstruct_swaps: bool
        remove_diagonal_gates_before_measure: bool
        reorder_operations: bool
        transform_dynamic_circuit: bool
        def __init__(self) -> None: ...

    class Parameterized:
        additional_instantiations: int
        parameterized_tolerance: float
        def __init__(self) -> None: ...

    class Simulation:
        fidelity_threshold: float
        max_sims: int
        seed: int
        state_type: StateType
        store_cex_input: bool
        store_cex_output: bool
        def __init__(self) -> None: ...
    application: Configuration.Application
    execution: Configuration.Execution
    functionality: Configuration.Functionality
    optimizations: Configuration.Optimizations
    parameterized: Configuration.Parameterized
    simulation: Configuration.Simulation
    def __init__(self) -> None: ...
    def json(self) -> dict[str, Any]: ...

class EquivalenceCheckingManager:
    class Results:
        cex_input: list[complex]
        cex_output1: list[complex]
        cex_output2: list[complex]
        check_time: float
        equivalence: EquivalenceCriterion
        performed_instantiations: int
        performed_simulations: int
        preprocessing_time: float
        started_simulations: int
        def __init__(self) -> None: ...
        def considered_equivalent(self) -> bool: ...
        def json(self) -> dict[str, Any]: ...

    def __init__(
        self, circ1: QuantumCircuit | str, circ2: QuantumCircuit | str, config: Configuration = ...
    ) -> None: ...
    def disable_all_checkers(self) -> None: ...
    def equivalence(self) -> EquivalenceCriterion: ...
    def fix_output_permutation_mismatch(self) -> None: ...
    def fuse_single_qubit_gates(self) -> None: ...
    def get_configuration(self) -> Configuration: ...
    def get_results(self) -> EquivalenceCheckingManager.Results: ...
    def json(self) -> dict[str, Any]: ...
    def reconstruct_swaps(self) -> None: ...
    def reorder_operations(self) -> None: ...
    def run(self) -> None: ...
    def set_alternating_application_scheme(self, scheme: ApplicationScheme | ApplicationSchemeName = ...) -> None: ...
    def set_alternating_checker(self, enable: bool = ...) -> None: ...
    def set_alternating_gate_cost_profile(self, profile: str = ...) -> None: ...
    def set_application_scheme(self, scheme: ApplicationScheme | ApplicationSchemeName = ...) -> None: ...
    def set_construction_application_scheme(self, scheme: ApplicationScheme | ApplicationSchemeName = ...) -> None: ...
    def set_construction_checker(self, enable: bool = ...) -> None: ...
    def set_construction_gate_cost_profile(self, profile: str = ...) -> None: ...
    def set_fidelity_threshold(self, threshold: float = ...) -> None: ...
    def set_gate_cost_profile(self, profile: str = ...) -> None: ...
    def set_max_sims(self, sims: int = ...) -> None: ...
    def set_nthreads(self, nthreads: int = ...) -> None: ...
    def set_parallel(self, enable: bool = ...) -> None: ...
    def set_seed(self, seed: int = ...) -> None: ...
    def set_simulation_application_scheme(self, scheme: ApplicationScheme | ApplicationSchemeName = ...) -> None: ...
    def set_simulation_checker(self, enable: bool = ...) -> None: ...
    def set_simulation_gate_cost_profile(self, profile: str = ...) -> None: ...
    def set_state_type(self, state_type: StateType | StateTypeName = ...) -> None: ...
    def set_timeout(self, timeout: float = ...) -> None: ...
    def set_tolerance(self, tolerance: float = ...) -> None: ...
    def set_trace_threshold(self, threshold: float = ...) -> None: ...
    def set_zx_checker(self, enable: bool = ...) -> None: ...
    def store_cex_input(self, enable: bool = ...) -> None: ...
    def store_cex_output(self, enable: bool = ...) -> None: ...

class EquivalenceCriterion:
    __members__: ClassVar[dict[EquivalenceCriterion, int]] = ...  # read-only
    equivalent: ClassVar[EquivalenceCriterion] = ...
    equivalent_up_to_global_phase: ClassVar[EquivalenceCriterion] = ...
    equivalent_up_to_phase: ClassVar[EquivalenceCriterion] = ...
    no_information: ClassVar[EquivalenceCriterion] = ...
    not_equivalent: ClassVar[EquivalenceCriterion] = ...
    probably_equivalent: ClassVar[EquivalenceCriterion] = ...
    probably_not_equivalent: ClassVar[EquivalenceCriterion] = ...
    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: EquivalenceCriterionName) -> None: ...
    def name(self) -> EquivalenceCriterionName: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> EquivalenceCriterionName: ...
    @property
    def value(self) -> int: ...

class StateType:
    __members__: ClassVar[dict[StateType, int]] = ...  # read-only
    computational_basis: ClassVar[StateType] = ...
    random_1Q_basis: ClassVar[StateType] = ...  # noqa: N815
    stabilizer: ClassVar[StateType] = ...
    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: StateTypeName) -> None: ...
    def name(self) -> StateTypeName: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> StateTypeName: ...
    @property
    def value(self) -> int: ...



"""MQT QCEC library.

This file is part of the MQT QCEC library released under the MIT license.
See README.md or go to https://github.com/cda-tum/qcec for more information.
"""
from __future__ import annotations

from ._version import version as __version__
from .compilation_flow_profiles import AncillaMode, generate_profile
from .pyqcec import (
    ApplicationScheme,
    Configuration,
    EquivalenceCheckingManager,
    EquivalenceCriterion,
    StateType,
)
from .verify import verify
from .verify_compilation_flow import verify_compilation

__all__ = [
    "AncillaMode",
    "ApplicationScheme",
    "Configuration",
    "EquivalenceCheckingManager",
    "EquivalenceCriterion",
    "generate_profile",
    "StateType",
    "verify",
    "verify_compilation",
    "__version__",
]



"""Types for the QCEC module."""
from __future__ import annotations

from typing import Literal

ApplicationSchemeName = Literal["sequential", "one_to_one", "proportional", "lookahead", "gate_cost"]
EquivalenceCriterionName = Literal[
    "no_information",
    "not_equivalent",
    "equivalent",
    "equivalent_up_to_phase",
    "equivalent_up_to_global_phase",
    "probably_equivalent",
    "probably_not_equivalent",
]
StateTypeName = Literal["computational_basis", "random_1Q_basis", "stabilizer"]



"""Configuration options for the equivalence checking manager."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from . import ApplicationScheme, Configuration, StateType
    from .types import ApplicationSchemeName, StateTypeName


class ConfigurationOptions(TypedDict, total=False):
    """A dictionary of configuration options.

    The keys of this dictionary are the names of the configuration options.
    The values are the values of the configuration options.
    """

    # Application
    alternating_scheme: ApplicationScheme | ApplicationSchemeName
    construction_scheme: ApplicationScheme | ApplicationSchemeName
    profile: str
    simulation_scheme: ApplicationScheme | ApplicationSchemeName
    # Execution
    nthreads: int
    numerical_tolerance: float
    parallel: bool
    run_alternating_checker: bool
    run_construction_checker: bool
    run_simulation_checker: bool
    run_zx_checker: bool
    timeout: float
    # Functionality
    trace_threshold: float
    # Optimizations
    fix_output_permutation_mismatch: bool
    fuse_single_qubit_gates: bool
    reconstruct_swaps: bool
    remove_diagonal_gates_before_measure: bool
    reorder_operations: bool
    transform_dynamic_circuit: bool
    # Parameterized
    additional_instantiations: int
    parameterized_tolerance: float
    # Simulation
    fidelity_threshold: float
    max_sims: int
    seed: int
    state_type: StateType | StateTypeName
    store_cex_input: bool
    store_cex_output: bool


def augment_config_from_kwargs(config: Configuration, kwargs: ConfigurationOptions) -> None:
    """Augment an existing :class:`~mqt.qcec.Configuration` with options from a collection of keyword arguments.

    Args:
        config: The configuration to augment.
        kwargs: The arguments to build the configuration from.
    """
    for key, value in kwargs.items():
        if hasattr(config.application, key):
            setattr(config.application, key, value)
        elif hasattr(config.execution, key):
            setattr(config.execution, key, value)
        elif hasattr(config.functionality, key):
            setattr(config.functionality, key, value)
        elif hasattr(config.optimizations, key):
            setattr(config.optimizations, key, value)
        elif hasattr(config.parameterized, key):
            setattr(config.parameterized, key, value)
        elif hasattr(config.simulation, key):
            setattr(config.simulation, key, value)
        else:  # pragma: no cover
            msg = f"Invalid keyword argument: {key}"
            raise ValueError(msg)



"""Functionality for checking equivalence of parameterized quantum circuits."""

from __future__ import annotations

import time
from itertools import chain
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from qiskit import QuantumCircuit

import numpy as np
from qiskit.circuit import Parameter, ParameterExpression

from . import Configuration, EquivalenceCheckingManager, EquivalenceCriterion


def __is_parameterized(qc: QuantumCircuit | str) -> bool:
    return not isinstance(qc, str) and qc.parameters


def __adjust_timeout(curr_timeout: float, res: EquivalenceCheckingManager.Results | float) -> float:
    if curr_timeout == 0:
        return 0

    if isinstance(res, EquivalenceCheckingManager.Results):
        return curr_timeout - (res.check_time + res.preprocessing_time)
    return curr_timeout - res


def check_parameterized_zx(
    circ1: QuantumCircuit | str, circ2: QuantumCircuit | str, configuration: Configuration
) -> EquivalenceCheckingManager.Results:
    """Check circuits for equivalence with the ZX-calculus."""
    ecm = EquivalenceCheckingManager(circ1, circ2, configuration)
    ecm.disable_all_checkers()
    ecm.set_zx_checker(True)

    ecm.run()

    return ecm.get_results()


def extract_params(
    circ1: QuantumCircuit, circ2: QuantumCircuit
) -> tuple[list[Parameter], NDArray[np.float64], NDArray[np.float64]]:
    """Extract parameters and equations of parameterized circuits."""
    p1 = set(circ1.parameters)
    p2 = set(circ2.parameters)

    p = p1.union(p2)

    n_params = len(p)
    exprs = list(chain(*[instr[0].params for instr in circ1.data + circ2.data if instr[0].params != []]))

    def is_expr(x: float | Parameter | ParameterExpression) -> bool:
        return isinstance(x, (Parameter, ParameterExpression))

    symb_params: list[Parameter | ParameterExpression] = [param for param in p if is_expr(param)]
    symb_params.sort(key=lambda param: param.name)
    symb_exprs = list(filter(is_expr, exprs))

    offsets = np.zeros(len(symb_exprs))
    for row, expr in enumerate(symb_exprs):
        zero_map = {param: 0 for param in expr.parameters}
        offsets[row] = -float(expr.bind(zero_map))

    equs = np.zeros((len(symb_exprs), n_params))

    for col, param in enumerate(symb_params):
        for row, expr in enumerate(symb_exprs):
            one_map = {p: 0 for p in expr.parameters}
            if param in expr.parameters:
                one_map[param] = 1
                val = float((expr + offsets[row]).bind(one_map))
                equs[row, col] = val

    return symb_params, equs, offsets


def check_instantiated(
    circ1: QuantumCircuit, circ2: QuantumCircuit, configuration: Configuration
) -> EquivalenceCheckingManager.Results:
    """Check circuits for equivalence with DD equivalence checker."""
    ecm = EquivalenceCheckingManager(circ1, circ2, configuration)
    ecm.set_zx_checker(False)

    ecm.run()

    return ecm.get_results()


def check_instantiated_random(
    circ1: QuantumCircuit, circ2: QuantumCircuit, params: list[Parameter], configuration: Configuration
) -> EquivalenceCheckingManager.Results:
    """Check whether circuits are equivalent for random instantiation of symbolic parameters."""
    param_map = {}
    rng = np.random.default_rng()
    for p in params:
        param_map[p] = rng.random() * 2 * np.pi

    circ1_inst = circ1.bind_parameters(param_map)
    circ2_inst = circ2.bind_parameters(param_map)

    return check_instantiated(circ1_inst, circ2_inst, configuration)


def check_parameterized(
    circ1: QuantumCircuit | str, circ2: QuantumCircuit | str, configuration: Configuration
) -> EquivalenceCheckingManager.Results:
    """Equivalence checking flow for parameterized circuit."""
    total_preprocessing_time = 0.0
    total_runtime = 0.0
    total_simulations_started = 0
    total_simulations_finished = 0

    def update_stats(res: EquivalenceCheckingManager.Results) -> None:
        nonlocal total_preprocessing_time
        nonlocal total_runtime
        nonlocal total_simulations_started
        nonlocal total_simulations_finished
        total_preprocessing_time += res.preprocessing_time
        total_runtime += res.check_time
        total_simulations_started += res.started_simulations
        total_simulations_finished += res.performed_simulations

    def write_stats(i: int, res: EquivalenceCheckingManager.Results) -> None:
        nonlocal total_preprocessing_time
        nonlocal total_runtime
        nonlocal total_simulations_started
        nonlocal total_simulations_finished
        res.check_time = total_runtime
        res.preprocessing_time = total_preprocessing_time
        res.started_simulations = total_simulations_started
        res.performed_simulations = total_simulations_finished
        res.performed_instantiations = i

    res = check_parameterized_zx(circ1, circ2, configuration)

    if res.considered_equivalent():
        return res

    update_stats(res)

    timeout = __adjust_timeout(configuration.execution.timeout, res)
    n_checks = configuration.parameterized.additional_instantiations
    tol = configuration.parameterized.parameterized_tolerance

    parameters, mat, offsets = extract_params(circ1, circ2)

    def instantiate_params(
        qc1: QuantumCircuit, qc2: QuantumCircuit, b: NDArray[np.float64]
    ) -> tuple[QuantumCircuit, QuantumCircuit, float]:
        start_time = time.time()
        mat_pinv = np.linalg.pinv(mat)
        x = np.dot(mat_pinv, b)
        param_map = {param: x[i] for i, param in enumerate(parameters)}
        qc1_bound = qc1.bind_parameters(param_map)
        qc2_bound = qc2.bind_parameters(param_map)

        def round_zero_params(qc: QuantumCircuit) -> QuantumCircuit:
            for instr in qc.data:
                params = instr[0].params
                instr[0].params = [float(x) for x in params]
                instr[0].params = [0 if np.abs(x) < tol else x for x in instr[0].params]
            return qc

        qc1_bound = round_zero_params(qc1_bound)
        qc2_bound = round_zero_params(qc2_bound)
        return qc1_bound, qc2_bound, time.time() - start_time

    def instantiate_params_zero(
        qc1: QuantumCircuit, qc2: QuantumCircuit
    ) -> tuple[QuantumCircuit, QuantumCircuit, float]:
        return instantiate_params(qc1, qc2, offsets)

    def instantiate_params_phases(
        qc1: QuantumCircuit, qc2: QuantumCircuit
    ) -> tuple[QuantumCircuit, QuantumCircuit, float]:
        phases = [0, np.pi, np.pi / 2, -np.pi / 2, np.pi / 4, -np.pi / 4]
        rng = np.random.default_rng()
        b = rng.choice(phases, size=len(offsets)) + offsets
        return instantiate_params(qc1, qc2, b)

    circ1_inst, circ2_inst, runtime = instantiate_params_zero(circ1, circ2)
    timeout = __adjust_timeout(timeout, runtime)

    if timeout < 0:
        write_stats(1, res)
        res.equivalence = EquivalenceCriterion.no_information
        return res

    res = check_instantiated(circ1_inst, circ2_inst, configuration)
    update_stats(res)
    if res.equivalence == EquivalenceCriterion.not_equivalent:
        write_stats(1, res)
        return res

    for i in range(n_checks):
        circ1_inst, circ2_inst, runtime = instantiate_params_phases(circ1, circ2)
        timeout = __adjust_timeout(timeout, runtime)
        res = check_instantiated(circ1_inst, circ2_inst, configuration)
        timeout = __adjust_timeout(timeout, res)

        if timeout < 0:
            write_stats(i + 2, res)
            res.equivalence = EquivalenceCriterion.no_information
            return res

        update_stats(res)

        if res.equivalence == EquivalenceCriterion.not_equivalent:
            write_stats(i + 2, res)
            return res

    res = check_instantiated_random(circ1, circ2, parameters, configuration)
    timeout = __adjust_timeout(timeout, runtime)
    if timeout < 0:
        res.equivalence = EquivalenceCriterion.no_information

    update_stats(res)
    write_stats(n_checks + 2, res)
    return res



"""Module for generating compilation flow profiles for the equivalence checking process."""

from __future__ import annotations

from enum import Enum, unique
from pathlib import Path
from typing import Any

import numpy as np
from qiskit import QuantumCircuit, transpile


@unique
class AncillaMode(Enum):
    """Enum for the ancilla mode."""

    NO_ANCILLA = "noancilla"
    RECURSION = "recursion"
    V_CHAIN = "v-chain"

    def __eq__(self, other: object) -> bool:
        """Check if two AncillaMode objects are equal. Supports string comparison."""
        if isinstance(other, str):
            return self.value == other
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False

    def __hash__(self) -> int:
        """Return the hash of the AncillaMode."""
        return hash(self.value)

    def __str__(self) -> str:
        """Return the string representation of the AncillaMode."""
        return self.value


single_qubit_gates_no_params = {
    "qubits": 1,
    "params": 0,
    "controls": 0,
    "gates": ["x", "y", "z", "h", "s", "sdg", "t", "tdg", "sx", "sxdg"],
}
single_qubit_gates_one_param = {
    "qubits": 1,
    "params": 1,
    "controls": 0,
    "gates": ["p", "rx", "ry", "rz"],
}
single_qubit_gates_three_params = {
    "qubits": 1,
    "params": 3,
    "controls": 0,
    "gates": ["u"],
}

two_qubit_gates = {"qubits": 2, "params": 0, "controls": 0, "gates": ["swap", "iswap"]}

single_controlled_single_qubit_gates_no_params = {
    "qubits": 1,
    "params": 0,
    "controls": 1,
    "gates": ["x", "y", "z", "h", "sx"],
}
single_controlled_single_qubit_gates_one_param = {
    "qubits": 1,
    "params": 1,
    "controls": 1,
    "gates": ["p", "rx", "ry", "rz"],
}
single_controlled_single_qubit_gates_three_params = {
    "qubits": 1,
    "params": 4,
    "controls": 1,
    "gates": ["u"],
}

single_controlled_two_qubit_gates = {
    "qubits": 2,
    "params": 0,
    "controls": 1,
    "gates": ["swap"],
}

double_controlled_single_qubit_gates_no_params = {
    "qubits": 1,
    "params": 0,
    "controls": 2,
    "gates": ["x"],
}

max_controls = 11
control_range = range(2, max_controls + 1)
mcx_no_ancilla = {
    "qubits": 1,
    "params": 0,
    "controls": control_range,
    "mode": AncillaMode.NO_ANCILLA,
    "ancilla_qubits": 0,
    "gates": ["x"],
}
mcx_recursion = {
    "qubits": 1,
    "params": 0,
    "controls": control_range,
    "mode": AncillaMode.RECURSION,
    "ancilla_qubits": 1,
    "gates": ["x"],
}
mcx_v_chain = {
    "qubits": 1,
    "params": 0,
    "controls": control_range,
    "mode": AncillaMode.V_CHAIN,
    "ancilla_qubits": None,
    "gates": ["x"],
}
mcphase = {
    "qubits": 1,
    "params": 1,
    "controls": control_range,
    "mode": None,
    "ancilla_qubits": 0,
    "gates": ["p"],
}

general_gates = [
    single_qubit_gates_no_params,
    single_qubit_gates_one_param,
    single_qubit_gates_three_params,
    two_qubit_gates,
    single_controlled_single_qubit_gates_no_params,
    single_controlled_single_qubit_gates_one_param,
    single_controlled_single_qubit_gates_three_params,
    single_controlled_two_qubit_gates,
    double_controlled_single_qubit_gates_no_params,
]

multi_controlled_gates = [mcphase]
multi_controlled_gates_no_ancilla = [mcx_no_ancilla]
multi_controlled_gates_recursion = [mcx_recursion]
multi_controlled_gates_v_chain = [mcx_v_chain]


def create_general_gate(qubits: int, params: int, controls: int, identifier: str) -> QuantumCircuit:
    """Create a ``QuantumCircuit`` containing a single gate ``identifier`` with the given number of ``qubits``, ``params``, and ``controls``."""
    required_qubits = qubits + controls
    qc = QuantumCircuit(required_qubits)
    gate_identifier = "c" * controls + identifier

    rng = np.random.default_rng()
    parameter_list = [rng.uniform(-np.pi, np.pi) for _ in range(params)]

    getattr(qc, gate_identifier)(*parameter_list, *range(required_qubits))
    return qc


def create_multi_controlled_gate(
    qubits: int,
    params: int,
    controls: int,
    mode: AncillaMode | None,
    ancilla_qubits: int | None,
    identifier: str,
) -> QuantumCircuit:
    """Create a ``QuantumCircuit`` containing a single multi-controlled gate ``identifier`` with the given number of ``qubits``, ``params``, and ``controls`` using ``ancilla_qubits`` ancilla qubits and the given ancilla ``mode``."""
    required_qubits = qubits + controls

    # special handling for v-chain mode which is indicated by the ancilla_qubits being None
    if ancilla_qubits is None:
        ancilla_qubits = max(0, controls - 2)

    # special handling for recursion mode with less than 5 controls,
    # which does not require ancilla qubits
    no_ancilla_threshold = 5
    if mode == "recursion" and controls < no_ancilla_threshold:
        ancilla_qubits = 0

    required_qubits += ancilla_qubits
    qc = QuantumCircuit(required_qubits)
    gate_identifier = "mc" + identifier

    rng = np.random.default_rng()
    parameter_list = [rng.uniform(-np.pi, np.pi) for _ in range(params)]

    if mode is not None:
        getattr(qc, gate_identifier)(
            *parameter_list,
            control_qubits=list(range(controls)),
            target_qubit=controls,
            ancilla_qubits=list(range(controls + 1, controls + 1 + ancilla_qubits)),
            mode=mode,
        )
    else:
        getattr(qc, gate_identifier)(*parameter_list, control_qubits=list(range(controls)), target_qubit=controls)
    return qc


def compute_cost(
    qc: QuantumCircuit,
    basis_gates: list[str],
    optimization_level: int = 1,
) -> int:
    """Compute the cost of a circuit by transpiling the circuit to a given ``basis_gates`` gate set and a certain ``optimization_level``."""
    transpiled_circuit = transpile(
        qc, basis_gates=basis_gates, optimization_level=optimization_level, seed_transpiler=12345
    )
    size: int = transpiled_circuit.size()
    return size


class GateType(Enum):
    """Enum for gate types."""

    GENERAL = 1
    MULTI_CONTROLLED = 2


def create_gate_profile_data(
    gate_collection: list[dict[str, Any]],
    gate_type: GateType,
    basis_gates: list[str] | None = None,
    optimization_level: int = 1,
) -> dict[tuple[str, int], int]:
    """Create a dictionary of gate profile data."""
    if basis_gates is None:
        basis_gates = ["id", "rz", "sx", "x", "cx"]

    profile_data = {}
    for gate_set in gate_collection:
        gates = gate_set["gates"]
        qubits = gate_set["qubits"]
        params = gate_set["params"]
        controls = gate_set["controls"]

        # pack single control numbers into list
        if gate_type == GateType.GENERAL:
            controls = [controls]

        for gate in gates:
            for control in controls:
                qc = None
                # create the gate
                if gate_type == GateType.GENERAL:
                    qc = create_general_gate(qubits, params, control, gate)
                elif gate_type == GateType.MULTI_CONTROLLED:
                    qc = create_multi_controlled_gate(
                        qubits,
                        params,
                        control,
                        gate_set["mode"],
                        gate_set["ancilla_qubits"],
                        gate,
                    )
                # compute the cost
                cost = compute_cost(qc, basis_gates, optimization_level)

                # add the cost to the profile data
                profile_data[(gate, control)] = cost
    return profile_data


def add_special_case_data(
    profile_data: dict[tuple[str, int], int],
    special_cases: dict[str, Any] | None = None,
) -> None:
    """Add special case data to the profile data. This is used to extrapolate the cost of specific rotation gates (e.g., S, T, ...) from the cost of the generic phase gate."""
    if special_cases is None:
        special_cases = {
            "gates": ["z", "s", "sdg", "t", "tdg"],
            "underlying_gate": "p",
        }

    gates = special_cases["gates"]
    underlying_gate = special_cases["underlying_gate"]

    for (g, nc), cost in profile_data.copy().items():
        if g is underlying_gate:
            for gate in gates:
                profile_data.setdefault((gate, nc), cost)


def generate_profile_name(optimization_level: int = 1, mode: AncillaMode = AncillaMode.NO_ANCILLA) -> str:
    """Generate a profile name based on the given optimization level and ancilla mode."""
    return "qiskit_O" + str(optimization_level) + "_" + str(mode) + ".profile"


def write_profile_data_to_file(profile_data: dict[tuple[str, int], int], filename: Path) -> None:
    """Write the profile data to a file."""
    with Path(filename).open("w+", encoding="utf-8") as f:
        for (gate, controls), cost in profile_data.items():
            f.write(f"{gate} {controls} {cost}\n")


def check_recurrence(seq: list[int], order: int = 2) -> list[int] | None:
    """Determine a recurrence relation with a given ``order`` in ``sequence`` and return the corresponding coefficients or ``None`` if no relation was determined."""
    if len(seq) < (2 * order + 1):
        return None

    mat, f = [], []
    for i in range(order):
        mat.append(seq[i : i + order])
        f.append(seq[i + order])

    if np.linalg.det(mat) == 0:
        return None

    coeffs = np.linalg.inv(mat).dot(f)

    for i in range(2 * order, len(seq)):
        predict = np.sum(coeffs * np.array(seq[i - order : i]))
        if abs(predict - seq[i]) > 10 ** (-2):
            return None

    return list(coeffs)


def find_continuation(
    profile_data: dict[tuple[str, int], int],
    gate: str,
    cutoff: int = 5,
    max_order: int = 3,
    max_control: int = 11,
    max_qubits: int = 128,
    prediction_cutoff: float = 1e6,
) -> None:
    """Extrapolate from the given profile data by finding recurrence relations."""
    sequence = [cost for (g, _), cost in profile_data.items() if g == gate]

    # sort the sequence
    sequence = sorted(sequence)

    # cut off the sequence at the cutoff point
    sequence = sequence[cutoff:]

    coeffs = None
    for order in range(1, max_order + 1):
        coeffs = check_recurrence(sequence, order)
        if coeffs is not None:
            break

    # if no recurrence sequence was found, assume the sequence is linear
    if coeffs is None:
        coeffs = [-1, 2]

    entries_to_compute = max_qubits - max_control - 1
    order = len(coeffs)
    for i in range(entries_to_compute):
        next_term = int(np.round(np.sum(coeffs * np.array(sequence[-order:]))))
        if next_term >= prediction_cutoff:
            break
        sequence.append(next_term)
        profile_data[(gate, max_control + i + 1)] = next_term


gate_collection_for_mode = {
    AncillaMode.NO_ANCILLA: multi_controlled_gates_no_ancilla,
    AncillaMode.RECURSION: multi_controlled_gates_recursion,
    AncillaMode.V_CHAIN: multi_controlled_gates_v_chain,
}
default_profile_path = Path(__file__).resolve().parent.joinpath("profiles")


def generate_profile(
    optimization_level: int = 1,
    mode: AncillaMode = AncillaMode.NO_ANCILLA,
    filepath: Path | None = None,
) -> None:
    """Generate a compilation flow profile for the given optimization level and ancilla mode.

    Args:
        optimization_level:
            The IBM Qiskit optimization level to use for the profile (0, 1, 2, or 3). Defaults to 1.
        mode:
            The `ancilla mode <.AncillaMode>` used for realizing multi-controlled Toffoli gates, as available in Qiskit.
            Defaults to ``AncillaMode.NO_ANCILLA``.
        filepath:
            The path to the directory where the profile should be stored.
            Defaults to the ``profiles`` directory in the ``mqt.qcec`` package.
    """
    if filepath is None:
        filepath = default_profile_path

    # generate general profile data
    profile = create_gate_profile_data(general_gates, GateType.GENERAL, optimization_level=optimization_level)

    # add multi-controlled gates
    profile.update(
        create_gate_profile_data(
            multi_controlled_gates,
            GateType.MULTI_CONTROLLED,
            optimization_level=optimization_level,
        )
    )
    find_continuation(profile, gate="p", max_control=max_controls)

    gate_collection = gate_collection_for_mode[mode]

    # add multi-controlled gates with specific mode
    profile.update(
        create_gate_profile_data(
            gate_collection,
            GateType.MULTI_CONTROLLED,
            optimization_level=optimization_level,
        )
    )
    find_continuation(profile, gate="x", max_control=max_controls)

    # add special case data
    add_special_case_data(profile)

    # write profile data to file
    filename = generate_profile_name(optimization_level=optimization_level, mode=mode)
    filepath = filepath.joinpath(filename)
    write_profile_data_to_file(profile, filepath)
    print(f"Wrote profile data to {filepath}")



"""Nox sessions."""

from __future__ import annotations

import argparse
import os
import sys
from typing import TYPE_CHECKING

import nox

if TYPE_CHECKING:
    from collections.abc import Sequence

nox.options.sessions = ["lint", "pylint", "tests"]

PYTHON_ALL_VERSIONS = ["3.8", "3.9", "3.10", "3.11", "3.12"]

if os.environ.get("CI", None):
    nox.options.error_on_missing_interpreters = True


@nox.session(reuse_venv=True)
def lint(session: nox.Session) -> None:
    """Lint the Python part of the codebase using pre-commit.

    Simply execute `nox -rs lint` to run all configured hooks.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session(reuse_venv=True)
def pylint(session: nox.Session) -> None:
    """Run PyLint.

    Simply execute `nox -rs pylint` to run PyLint.
    """
    session.install("scikit-build-core[pyproject]", "setuptools_scm", "pybind11")
    session.install("--no-build-isolation", "-ve.", "pylint")
    session.run("pylint", "mqt.qcec", *session.posargs)


def _run_tests(
    session: nox.Session,
    *,
    install_args: Sequence[str] = (),
    run_args: Sequence[str] = (),
    extras: Sequence[str] = (),
) -> None:
    posargs = list(session.posargs)
    env = {"PIP_DISABLE_PIP_VERSION_CHECK": "1"}

    if os.environ.get("CI", None) and sys.platform == "win32":
        env["SKBUILD_CMAKE_ARGS"] = "-T ClangCL"

    _extras = ["test", *extras]
    if "--cov" in posargs:
        _extras.append("coverage")
        posargs.append("--cov-config=pyproject.toml")

    session.install("scikit-build-core[pyproject]", "setuptools_scm", "pybind11", *install_args, env=env)
    install_arg = f"-ve.[{','.join(_extras)}]"
    session.install("--no-build-isolation", install_arg, *install_args, env=env)
    session.run("pytest", *run_args, *posargs, env=env)


@nox.session(reuse_venv=True, python=PYTHON_ALL_VERSIONS)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    _run_tests(session)


@nox.session()
def minimums(session: nox.Session) -> None:
    """Test the minimum versions of dependencies."""
    _run_tests(
        session,
        install_args=["--constraint=test/python/constraints.txt"],
        run_args=["-Wdefault"],
    )
    session.run("pip", "list")


@nox.session(reuse_venv=True)
def docs(session: nox.Session) -> None:
    """Build the docs. Pass "--serve" to serve. Pass "-b linkcheck" to check links."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--serve", action="store_true", help="Serve after building")
    parser.add_argument("-b", dest="builder", default="html", help="Build target (default: html)")
    args, posargs = parser.parse_known_args(session.posargs)

    if args.builder != "html" and args.serve:
        session.error("Must not specify non-HTML builder with --serve")

    build_requirements = ["scikit-build-core[pyproject]", "setuptools_scm", "pybind11"]
    extra_installs = ["sphinx-autobuild"] if args.serve else []
    session.install(*build_requirements, *extra_installs)
    session.install("--no-build-isolation", "-ve.[docs]")
    session.chdir("docs")

    if args.builder == "linkcheck":
        session.run("sphinx-build", "-b", "linkcheck", "source", "_build/linkcheck", *posargs)
        return

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        "source",
        f"_build/{args.builder}",
        *posargs,
    )

    if args.serve:
        session.run("sphinx-autobuild", *shared_args)
    else:
        session.run("sphinx-build", "--keep-going", *shared_args)



"""Verify compilation flow results."""

from __future__ import annotations

import sys
import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from .configuration import ConfigurationOptions

if TYPE_CHECKING or sys.version_info < (3, 10, 0):
    import importlib_resources as resources
else:
    from importlib import resources

from qiskit import QuantumCircuit

from . import ApplicationScheme, Configuration, EquivalenceCheckingManager
from .compilation_flow_profiles import AncillaMode, generate_profile_name
from .configuration import augment_config_from_kwargs
from .verify import verify


def __check_if_circuit_contains_measurements(circuit: QuantumCircuit) -> None:
    """Check if the circuit contains measurements and emits a warning if it does not.

    Args:
        circuit: The circuit to check.
    """
    from qiskit.transpiler.passes import ContainsInstruction

    analysis_pass = ContainsInstruction("measure")
    analysis_pass(circuit)
    if not analysis_pass.property_set["contains_measure"]:
        warnings.warn(
            UserWarning(
                "One of the circuits does not contain any measurements. "
                "This may lead to unexpected results since the measurements are used "
                "to infer the output qubit permutation at the end of the circuit. "
                "Please consider adding measurements to the circuit _before_ compilation."
            ),
            stacklevel=2,
        )


def verify_compilation(
    original_circuit: QuantumCircuit | str,
    compiled_circuit: QuantumCircuit | str,
    optimization_level: int = 1,
    ancilla_mode: AncillaMode = AncillaMode.NO_ANCILLA,
    configuration: Configuration | None = None,
    **kwargs: Unpack[ConfigurationOptions],
) -> EquivalenceCheckingManager.Results:
    """Verify compilation flow results.

    Similar to :func:`verify <.verify>`, but uses a dedicated compilation flow profile to guide the equivalence checking process.
    The compilation flow profile is determined by the ``optimization_level`` and ``ancilla_mode`` arguments.

    There are two (non-exclusive) ways of configuring the equivalence checking process:

    1. Pass a :class:`Configuration <.Configuration>` instance as the ``configuration`` argument.

    2. Pass keyword arguments to this function. These are directly incorporated into the :class:`Configuration <.Configuration>`.
    Any existing configuration is overridden by keyword arguments.

    Args:
        original_circuit: The original circuit.
        compiled_circuit: The compiled circuit.
        optimization_level: The optimization level used for compiling the circuit (0, 1, 2, or 3). Defaults to 1.
        ancilla_mode:
            The `ancilla mode <.AncillaMode>` used for realizing multi-controlled Toffoli gates, as available in Qiskit.
            Defaults to :attr:`.AncillaMode.NO_ANCILLA`.
        configuration: The configuration to use for the equivalence checking process.
        **kwargs: Keyword arguments to configure the equivalence checking process.

    Returns:
        The results of the equivalence checking process.
    """
    if configuration is None:
        configuration = Configuration()

    # prepare the configuration
    augment_config_from_kwargs(configuration, kwargs)

    if isinstance(original_circuit, QuantumCircuit):
        __check_if_circuit_contains_measurements(original_circuit)

    if isinstance(compiled_circuit, QuantumCircuit):
        __check_if_circuit_contains_measurements(compiled_circuit)

    # use the gate_cost scheme for the verification
    configuration.application.construction_scheme = ApplicationScheme.gate_cost
    configuration.application.simulation_scheme = ApplicationScheme.gate_cost
    configuration.application.alternating_scheme = ApplicationScheme.gate_cost

    # get the pre-defined profile for the gate_cost scheme
    profile_name = generate_profile_name(optimization_level=optimization_level, mode=ancilla_mode)
    ref = resources.files("mqt.qcec") / "profiles" / profile_name
    with resources.as_file(ref) as path:
        configuration.application.profile = str(path)

    return verify(original_circuit, compiled_circuit, configuration=configuration)



"""The main entry point for the QCEC package."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qiskit import QuantumCircuit
    from typing_extensions import Unpack

    from .configuration import ConfigurationOptions

from . import Configuration, EquivalenceCheckingManager
from .configuration import augment_config_from_kwargs
from .parameterized import __is_parameterized, check_parameterized


def verify(
    circ1: QuantumCircuit | str,
    circ2: QuantumCircuit | str,
    configuration: Configuration | None = None,
    **kwargs: Unpack[ConfigurationOptions],
) -> EquivalenceCheckingManager.Results:
    """Verify that ``circ1`` and ``circ2`` are equivalent.

    Wraps creating an instance of :class:`EquivalenceCheckingManager <.EquivalenceCheckingManager>`,
    calling :meth:`EquivalenceCheckingManager.run`,
    and calling :meth:`EquivalenceCheckingManager.get_results`.

    There are two (non-exclusive) ways of configuring the equivalence checking process:

    1. Pass a :class:`Configuration <.Configuration>` instance as the ``configuration`` argument.

    2. Pass keyword arguments to this function. These are directly incorporated into the :class:`Configuration <.Configuration>`.
    Any existing configuration is overridden by keyword arguments.

    Args:
        circ1: The first circuit.
        circ2: The second circuit.
        configuration: The configuration to use for the equivalence checking process.
        **kwargs: Keyword arguments to configure the equivalence checking process.

    Returns:
        The results of the equivalence checking process.
    """
    if configuration is None:
        configuration = Configuration()

    # prepare the configuration
    augment_config_from_kwargs(configuration, kwargs)

    if __is_parameterized(circ1) or __is_parameterized(circ2):
        return check_parameterized(circ1, circ2, configuration)

    # create the equivalence checker from configuration
    ecm = EquivalenceCheckingManager(circ1, circ2, configuration)

    # execute the check
    ecm.run()

    # obtain the result
    return ecm.get_results()
